@@ -6,49 +6,14 @@ import contextlib
 import zipfile
 import h5py
 import uuid
+import logging
 
 from functools import lru_cache
-from slugify import slugify
 from pathlib import Path
+from cached_property import cached_property
 from vectorian.embeddings import MaskedVectorsRef, ExternalMemoryVectorsRef
 
 
-class Text:
-	def get(self):
-		raise NotImplementedError()
-
-	def close(self):
-		raise NotImplementedError()
-
-
-class InternalMemoryText(Text):
-	def __init__(self, text):
-		self._text = text
-
-	def get(self):
-		return self._text
-
-	def close(self):
-		pass
-
-
-class ExternalMemoryText(Text):
-	def __init__(self, path):
-		self._f = open(path, "r")
-		self._text = self._f.read()
-		# FIXME
-
-	def get(self):
-		return self._text
-
-	def close(self):
-		self._f.close()
-
-
-def convert_idx_len_to_utf8(text, idx, len):
-	pass  # FIXME
-
-
 class TokenTable:
 	def __init__(self, text, normalizers):
 		self._text = text
@@ -59,6 +24,7 @@ class TokenTable:
 		self._token_tag = []  # tag_ from spacy's Token
 		self._token_str = []
 
+		self._normalizers = normalizers
 		self._make_text = normalizers['token'].token_to_text
 		self._norm_text = normalizers['text'].to_callable()
 
@@ -93,6 +59,55 @@ class TokenTable:
 		}
 
 
+class Text:
+	def get(self):
+		raise NotImplementedError()
+
+	def close(self):
+		raise NotImplementedError()
+
+
+class InternalMemoryText(Text):
+	def __init__(self, text):
+		self._text = text
+
+	def get(self):
+		return self._text
+
+	def close(self):
+		pass
+
+
+class ExternalMemoryText(Text):
+	def __init__(self, path):
+		self._f = open(path, "r")
+		self._text = self._f.read()
+
+	def get(self):
+		return self._text
+
+	def close(self):
+		self._f.close()
+		
+		
+class ExternalSqliteText(Text):
+	def __init__(self, doc_sql, unique_id):
+		self._doc_sql = doc_sql
+		self._unique_id = unique_id
+
+	def get(self):
+		cursor = self._doc_sql.cursor()
+		try:
+			cursor.execute("SELECT content FROM text WHERE unique_id=?", (self._unique_id,))
+			text = cursor.fetchone()[0]
+		finally:
+			cursor.close()
+		return text
+
+	def close(self):
+		pass
+
+
 class Lengths:
 	def __init__(self, start, end):
 		self._start = start
@@ -122,6 +137,26 @@ class DocumentStorage:
 		raise NotImplementedError()
 
 
+class Table:
+	def __init__(self, get, count):
+		self._col = {}
+		self._get = get
+		self._count = count
+
+	def __len__(self):
+		return self._count
+
+	def __getitem__(self, k):
+		data = self._col.get(k)
+		if data is None:
+			data = self._get(k)
+			self._col[k] = data
+		return data
+
+	def __setitem__(self, k, v):
+		self._col[k] = v
+
+
 class Tokens:
 	pass
 
@@ -154,6 +189,9 @@ class InternalMemoryTokens(Tokens):
 		for i in range(len(self)):
 			yield InternalMemoryTokens.TokenData(self._data, i)
 
+	def to_table(self):
+		return Table(self._data.get, len(self))
+
 	def save_to_h5(self, hf):
 		for k, v in self._tokens.items():
 			dtype = v['dtype']
@@ -194,8 +232,8 @@ class ExternalMemoryTokens(Tokens):
 		def __getitem__(self, key):
 			return self.get(key)
 
-	def __init__(self, path):
-		self._hf = h5py.File(path, 'r')
+	def __init__(self, hf):
+		self._hf = hf
 		self._cache = dict()
 
 	def _column(self, k):
@@ -228,27 +266,23 @@ class ExternalMemoryTokens(Tokens):
 		for i in range(len(self)):
 			yield ExternalMemoryTokens.TokenData(self._hf, get, i)
 
-	def close(self):
-		self._hf.close()
+	def to_table(self):
+		return Table(self._column, len(self))
 
 
 class InternalMemoryDocumentStorage(DocumentStorage):
-	def __init__(self, json, text, tokens, spans):
-		self._json = json
+	def __init__(self, metadata, text, tokens, spans):
+		self._metadata = metadata
 		self._text = text
 		self._tokens = InternalMemoryTokens(tokens)
 		self._spans = spans
 
 	@property
 	def metadata(self):
-		return self._json['metadata']
+		return self._metadata
 
 	@contextlib.contextmanager
-	def json(self):
-		yield self._json
-
-	@contextlib.contextmanager
-	def text(self, session):
+	def text(self, cache=None):
 		yield InternalMemoryText(self._text)
 
 	@contextlib.contextmanager
@@ -260,33 +294,60 @@ class InternalMemoryDocumentStorage(DocumentStorage):
 		yield self._spans
 
 
+def _spans_from_h5(hf):
+	spans = dict()
+	for name in hf.keys():
+		data = dict()
+		for k, v in hf[name].items():
+			data[k] = v
+		spans[name] = data
+
+	return spans
+
+
+class CorpusDocumentStorage(DocumentStorage):
+	def __init__(self, doc_path, doc_sql, doc_group, unique_id):
+		self._doc_path = doc_path
+		self._doc_sql = doc_sql
+		self._doc_group = doc_group
+		self._unique_id = unique_id
+
+	@cached_property
+	def metadata(self):
+		return json.loads(self._doc_group.attrs["metadata"])
+
+	@contextlib.contextmanager
+	def text(self, cache=None):
+		if self._doc_sql is None:
+			text = ExternalMemoryText(self._doc_path / "document.txt")
+			try:
+				yield text
+			finally:
+				text.close()
+		else:
+			yield ExternalSqliteText(self._doc_sql, self._unique_id)
+
+	@contextlib.contextmanager
+	def tokens(self):
+		yield ExternalMemoryTokens(self._doc_group["tokens"])
+
+	@contextlib.contextmanager
+	def spans(self):
+		yield _spans_from_h5(self._doc_group["spans"])
+
+
 class ExternalMemoryDocumentStorage(DocumentStorage):
 	def __init__(self, path):
 		self._path = Path(path)
-		self._metadata = None
 
-	def _load_metadata(self, zf):
-		data = json.loads(zf.read('data.json'))
-		self._metadata = data['metadata']
-		self._metadata['origin'] = str(self._path)
-
-	@property
+	@cached_property
 	def metadata(self):
-		if self._metadata is None:
-			with zipfile.ZipFile(self._path.with_suffix('.zip'), 'r') as zf:
-				self._load_metadata(zf)
-		return self._metadata
+		with open("metadata.json", "r") as f:
+			return json.loads(f.read())
 
 	@contextlib.contextmanager
-	def json(self):
-		with zipfile.ZipFile(self._path.with_suffix('.zip'), 'r') as zf:
-			if self._metadata is None:
-				self._load_metadata(zf)
-			yield json.loads(zf.read('data.json'))
-
-	@contextlib.contextmanager
-	def text(self, session):
-		text = ExternalMemoryText(self._path.with_suffix('.txt'))
+	def text(self, cache=None):
+		text = ExternalMemoryText(self._path / "document.txt")
 		try:
 			yield text
 		finally:
@@ -294,23 +355,13 @@ class ExternalMemoryDocumentStorage(DocumentStorage):
 
 	@contextlib.contextmanager
 	def tokens(self):
-		tokens = ExternalMemoryTokens(self._path.with_suffix(".tok.h5"))
-
-		try:
-			yield tokens
-		finally:
-			tokens.close()
+		with h5py.File(self._path / "tokens.h5", 'r') as hf:
+			yield ExternalMemoryTokens(hf)
 
 	@contextlib.contextmanager
 	def spans(self):
-		with h5py.File(self._path.with_suffix(".spn.h5"), "r") as hf:
-			spans = dict()
-			for name in hf.keys():
-				data = dict()
-				for k, v in hf[name].items():
-					data[k] = v
-				spans[name] = data
-			yield spans
+		with h5py.File(self._path / "spans.h5", "r") as hf:
+			yield _spans_from_h5(hf)
 
 
 class Document:
@@ -330,11 +381,79 @@ class Document:
 		return name in self._contextual_embeddings
 
 	@staticmethod
-	def load(path):
+	def load_from_corpus(unique_id, doc_path, doc_sql, doc_group):
+		contextual_embeddings = Document._load_embeddings(doc_path)
+		return Document(CorpusDocumentStorage(
+			doc_path, doc_sql, doc_group, unique_id), contextual_embeddings)
+
+	@staticmethod
+	def load_from_fs(path):
+		path = Path(path)
+		if path.suffix != ".document":
+			raise ValueError(f"document path '{path}' must end in '.document'")
+
+		contextual_embeddings = Document._load_embeddings(path)
+		return Document(ExternalMemoryDocumentStorage(path), contextual_embeddings)
+
+	def save_to_corpus(self, unique_id, doc_path, doc_sql, doc_group):
+		doc_group.attrs["metadata"] = json.dumps(self._storage.metadata)
+
+		with self._storage.tokens() as tokens:
+			tokens.save_to_h5(doc_group.create_group("tokens"))
+
+		with self._storage.spans() as spans:
+			spans_group = doc_group.create_group("spans")
+			for name, data in spans.items():
+				g = spans_group.create_group(name)
+				for k, v in data.items():
+					g.create_dataset(k, data=v)
+
+		with self._storage.text() as text:
+			if doc_sql is None:
+				doc_path.mkdir(exist_ok=True)
+
+				with open(doc_path / "document.txt", "w") as f:
+					f.write(text.get())
+			else:
+				with doc_sql:
+					doc_sql.execute('''
+						INSERT INTO text(unique_id, content) VALUES (?, ?)
+						''', (unique_id, text.get()))
+
+		self._save_embeddings(doc_path)
+
+	def save_to_fs(self, path):
 		path = Path(path)
 
+		if not path.is_dir():
+			raise ValueError(f"document path '{path}' needs to be a dir")
+		if path.suffix != ".document":
+			raise ValueError(f"document path '{path}' must end in '.document'")
+
+		with open(path / "metadata.json", "w") as f:
+			f.write(json.dumps(self._storage.metadata))
+
+		with self._storage.tokens() as tokens:
+			with h5py.File(path / "tokens.h5", "w") as hf:
+				tokens.save_to_h5(hf)
+
+		with self._storage.spans() as spans:
+			with h5py.File(path / "spans.h5", "w") as hf:
+				for name, data in spans.items():
+					g = hf.create_group(name)
+					for k, v in data.items():
+						g.create_dataset(k, data=v)
+
+		with self._storage.text() as text:
+			with open(path / "document.txt", "w") as f:
+				f.write(text.get())
+
+		self._save_embeddings(path)
+
+	@staticmethod
+	def _load_embeddings(path):
 		contextual_embeddings = dict()
-		emb_path = path.with_suffix(".embeddings")
+		emb_path = path / "embeddings"
 		emb_json_path = emb_path / "info.json"
 		if emb_json_path.exists():
 			with open(emb_json_path, "r") as f:
@@ -342,36 +461,14 @@ class Document:
 
 			for k, slug_name in emb_data.items():
 				contextual_embeddings[k] = ExternalMemoryVectorsRef(emb_path / slug_name)
+		return contextual_embeddings
 
-		return Document(ExternalMemoryDocumentStorage(path), contextual_embeddings)
-
-	def save(self, path):
-		path = Path(path)
-
-		with self._storage.json() as data:
-			with zipfile.ZipFile(path.with_suffix(".zip"), "w") as zf:
-				zf.writestr("data.json", json.dumps(data))
-
-		with self._storage.tokens() as tokens:
-			with h5py.File(path.with_suffix(".tok.h5"), "w") as hf:
-				tokens.save_to_h5(hf)
-
-		with self._storage.spans() as spans:
-			with h5py.File(path.with_suffix(".spn.h5"), "w") as hf:
-				for name, data in spans.items():
-					g = hf.create_group(name)
-					for k, v in data.items():
-						g.create_dataset(k, data=v)
-
-		with self._storage.text(None) as text:
-			with open(path.with_suffix(".txt"), "w") as f:
-				f.write(text.get())
-
+	def _save_embeddings(self, path):
 		if self._contextual_embeddings:
 			emb_data = dict()
 
-			emb_path = path.with_suffix(".embeddings")
-			emb_path.mkdir(exist_ok=True)
+			emb_path = path / "embeddings"
+			emb_path.mkdir(exist_ok=True, parents=True)
 
 			for k, vectors in self._contextual_embeddings.items():
 				unique_name = uuid.uuid1().hex
@@ -383,34 +480,26 @@ class Document:
 				f.write(json.dumps(emb_data))
 
 	def to_json(self):
-		with self._storage.json() as data:
-			r = data
-		return r
+		return self._storage.metadata
 
 	@property
 	def structure(self):
-		with self._storage.json() as data:
-			lines = []
-			for i, p in enumerate(data["partitions"]):
-				text = p["text"]
-				lines.append(f"partition {i + 1}:")
-				for j, sent in enumerate(p["sents"]):
-					lines.append(f"  sentence {j + 1}:")
-					lines.append("    " + text[sent["start"]:sent["end"]])
-			return "\n".join(lines)
+		data = self._storage.metadata
+
+		lines = []
+		for i, p in enumerate(data["partitions"]):
+			text = p["text"]
+			lines.append(f"partition {i + 1}:")
+			for j, sent in enumerate(p["sents"]):
+				lines.append(f"  sentence {j + 1}:")
+				lines.append("    " + text[sent["start"]:sent["end"]])
+
+		return "\n".join(lines)
 
 	@property
 	def metadata(self):
 		return self._storage.metadata
 
-	@property
-	def unique_id(self):
-		return self.metadata['unique_id']
-
-	@property
-	def caching_name(self):
-		return slugify(self.unique_id)
-
 	@property
 	def origin(self):
 		return self.metadata['origin']
@@ -419,12 +508,16 @@ class Document:
 	def title(self):
 		return self.metadata['title']
 
-	def prepare(self, session, doc_index):
-		names = [v.factory.name for v in session.embeddings.values() if v.factory.is_contextual]
-		contextual_embeddings = dict((k, self._contextual_embeddings[k]) for k in names)
+	def prepare(self, corpus, flavor_cache, session):
+		try:
+			names = [v.factory.name for v in session.embeddings.values() if v.factory.is_contextual]
+			contextual_embeddings = dict((k, self._contextual_embeddings[k]) for k in names)
 
-		return PreparedDocument(
-			self, session, doc_index, contextual_embeddings)
+			return PreparedDocument(
+				corpus, flavor_cache, session, self, contextual_embeddings)
+		except:
+			logging.error(f"failed to prepare doc '{corpus.get_unique_id(self)}'")
+			raise
 
 
 class Token:
@@ -509,24 +602,17 @@ class Span:
 
 
 class PreparedDocument:
-	def __init__(self, doc, session, doc_index, contextual_embeddings):
+	def __init__(self, corpus, flavor_cache, session, doc, contextual_embeddings):
 		self._doc = doc
-		storage = doc.storage
-
 		self._session = session
 
-		token_mapper = session.normalizer('token').token_to_token
+		storage = doc.storage
+		self._metadata = storage.metadata
 
-		with storage.tokens() as tokens:
-			with storage.text(self._session) as text:
-				token_table = TokenTable(text.get(), self._session.normalizers)
+		uid = corpus.get_unique_id(doc)
+		flavor_record = flavor_cache.get(uid)
 
-				token_mask = np.zeros((len(tokens),), dtype=np.bool)
-
-				for i, token in enumerate(tokens):
-					t = token_mapper(token)
-					if t:
-						token_mask[i] = token_table.add(t)
+		token_mask = flavor_record.token_mask
 
 		with storage.spans() as spans:
 			reindex = np.cumsum(np.concatenate(([False], token_mask)), dtype=np.int32)
@@ -541,13 +627,11 @@ class PreparedDocument:
 		self._contextual_embeddings = dict(
 			(k, MaskedVectorsRef(v, token_mask)) for k, v in contextual_embeddings.items())
 
-		self._metadata = storage.metadata
-
 		self._compiled = core.Document(
-			doc_index,
+			corpus.get_doc_index(doc),
 			session.vocab,
 			self._spans,
-			token_table.to_dict(),
+			flavor_record.to_dict(),
 			self._metadata,
 			self._contextual_embeddings)
 
@@ -586,15 +670,11 @@ class PreparedDocument:
 
 	@property
 	def unique_id(self):
-		return self.metadata['unique_id']
-
-	@property
-	def caching_name(self):
-		return slugify(self.unique_id)
+		return self._doc.unique_id
 
 	@contextlib.contextmanager
 	def text(self):
-		with self.storage.text(self._session) as text:
+		with self.storage.text() as text:
 			yield text
 
 	def token(self, i):
