@@ -1,6 +1,5 @@
 import re
 import logging
-import datetime
 import numpy as np
 import download
 
@@ -43,10 +42,14 @@ def compile_spans(spans, tokens, loc_ax):
 		'end': np.empty(n, dtype=np.int32)
 	}
 
-	loc = np.array(spans['loc'])
-	for i, k in enumerate(loc_ax):
-		assert k not in new_spans
-		new_spans[k] = to_min_dtype(loc[:, i])
+	if len(spans['loc']) > 0:
+		loc = np.array(spans['loc'])
+		for i, k in enumerate(loc_ax):
+			assert k not in new_spans
+			new_spans[k] = to_min_dtype(loc[:, i])
+	else:
+		for i, k in enumerate(loc_ax):
+			new_spans[k] = np.array([], dtype=np.uint8)
 
 	token_i = 0
 
@@ -101,8 +104,18 @@ def make_tokens_dict(tokens):
 	return res
 
 
+def get_text_from_spec(spec):
+	if isinstance(spec, Path):
+		with open(spec, "r") as f:
+			return f.read(), spec.stem, str(spec)
+	elif isinstance(spec, str):
+		return spec, "", "<string>"
+	else:
+		raise ValueError(f"unknown text specification {spec}")
+
+
 Metadata = namedtuple(
-	"Metadata", ["version", "unique_id", "origin", "author", "title", "speakers"])
+	"Metadata", ["version", "origin", "author", "title", "locations"])
 
 
 class Importer:
@@ -135,6 +148,7 @@ class Importer:
 		return text
 
 	def _make_doc(self, md, partitions, loc_ax, locations, show_progress=True):
+
 		pipe = self._nlp.pipe(
 			partitions,
 			batch_size=self._batch_size,
@@ -177,15 +191,17 @@ class Importer:
 			for e in self._embeddings:
 				contextual_vectors[e.name].append(e.encode(doc))
 
+		if not tokens:
+			return None
+
 		spans = {
 			'sentence': compile_spans(sents, tokens, loc_ax),
 			'document': compile_doc_spans(tokens)
 		}
 
-		json = {
-			'metadata': md._asdict(),
-			'loc_ax': loc_ax
-		}
+		extended_metadata = dict(
+			**md._asdict(),
+			loc_ax=loc_ax)
 
 		from vectorian.corpus.document import Document, InternalMemoryDocumentStorage
 
@@ -204,24 +220,27 @@ class Importer:
 			(k, transformed(k, v)) for k, v in contextual_vectors.items())
 
 		return Document(
-			InternalMemoryDocumentStorage(json, ''.join(texts), make_tokens_dict(tokens), spans),
+			InternalMemoryDocumentStorage(
+				extended_metadata, ''.join(texts), make_tokens_dict(tokens), spans),
 			contextual_embeddings)
 
+	def from_path(self, path: Path, *args, **kwargs):
+		raise NotImplementedError()
+
+	def from_str(self, text: str, *args, **kwargs):
+		raise NotImplementedError()
+
 
 class TextImporter(Importer):
 	# an importer for plain text files that does not assume any structure.
 
-	def __call__(self, path, unique_id=None, author="Anonymous", title=None):
-		path = Path(path)
+	def __call__(self, spec, author="", title=None, **kwargs):
+		text_base, title_base, origin = get_text_from_spec(spec)
 
 		if title is None:
-			title = path.stem
+			title = title_base
 
-		if unique_id is None:
-			unique_id = f"{author}/{title}"
-
-		with open(path, "r") as f:
-			text = self._preprocess_text(f.read())
+		text = self._preprocess_text(text_base)
 
 		locations = []
 
@@ -229,19 +248,25 @@ class TextImporter(Importer):
 		paragraphs = text.split(paragraph_sep)
 		paragraphs = [x + paragraph_sep for x in paragraphs[:-1]] + paragraphs[-1:]
 
+		if not paragraphs:
+			return None
+
 		for j, p in enumerate(paragraphs):
 			locations.append((j,))
 
 		md = Metadata(
 			version="1.0",
-			unique_id=unique_id,
-			origin=str(path),
+			origin=origin,
 			author=author,
 			title=title,
-			speakers={})
+			locations={
+				'type': 'text',
+				'data': {
+				}
+			})
 
 		return self._make_doc(
-			md, paragraphs, ['paragraph'], locations)
+			md, paragraphs, ['paragraph'], locations, **kwargs)
 
 
 class NovelImporter(Importer):
@@ -250,15 +275,12 @@ class NovelImporter(Importer):
 	_chapters = re.compile(
 		r"\n\n\n\W*chapter\s+(\d+)[^\n]*\n\n", re.IGNORECASE)
 
-	def __call__(self, path, unique_id=None, author="Anonymous", title=None):
+	def __call__(self, path, author="Anonymous", title=None, **kwargs):
 		path = Path(path)
 
 		if title is None:
 			title = path.stem
 
-		if unique_id is None:
-			unique_id = f"{author}/{title}"
-
 		with open(path, "r") as f:
 			text = self._preprocess_text(f.read())
 
@@ -317,14 +339,13 @@ class NovelImporter(Importer):
 
 		md = Metadata(
 			version="1.0",
-			unique_id=unique_id,
 			origin=str(path),
 			author=author,
 			title=title,
-			speakers={})
+			locations={'type': 'book'})
 
 		return self._make_doc(
-			md, paragraphs, ['book', 'chapter', 'paragraph'], locations)
+			md, paragraphs, ['book', 'chapter', 'paragraph'], locations, **kwargs)
 
 
 class BodleianImporter(Importer):
@@ -347,7 +368,7 @@ class PlayShakespeareImporter(Importer):
 
 		return self(xml_path)
 
-	def __call__(self, path):
+	def __call__(self, path, **kwargs):
 		import xml.etree.ElementTree as ET
 		from collections import defaultdict
 
@@ -389,38 +410,84 @@ class PlayShakespeareImporter(Importer):
 					locations.append((actnum, scenenum, speaker_no, line_no))
 					texts.append(normalize_dashes(" ".join(lines)))
 
-		unique_name = f"{root.attrib['variant']}-{root.attrib['unique']}"
-
 		md = Metadata(
 			version="1.0",
-			unique_id="PlayShakespeare.com/William Shakespeare/play/" + unique_name,
 			origin=str(path),
 			author="William Shakespeare",
 			title=root.find(".//title").text,
-			speakers={v: full_speaker_names.get(k, k) for k, v in speakers.items()})
+			locations={
+				'type': 'play',
+				'data': {
+					'speakers': {v: full_speaker_names.get(k, k) for k, v in speakers.items()}
+				}
+			})
 
 		return self._make_doc(
-			md, texts, ['act', 'scene', 'speaker', 'line'], locations)
+			md, texts, ['act', 'scene', 'speaker', 'line'], locations, **kwargs)
 
 
-class StringImporter(Importer):
-	# a generic importer for short text strings for ad-hoc experiments.
+class MarkdownImporter(Importer):
+	# a generic importer for markdown texts.
 
-	def __call__(self, s, unique_id=None, author="", title="", show_progress=False):
-		if unique_id is None:
-			unique_id = str(datetime.datetime.now())
+	_sections = re.compile(
+		r"\n#([^\n]+)\n", re.IGNORECASE)
 
-		locations = [
-			[]
-		]
+	def __call__(self, text, author="", title=None, **kwargs):
+
+		text_base, title_base, origin = get_text_from_spec(text)
+
+		if title is None:
+			title = title_base
+
+		text = "\n" + text_base
+
+		heading_breaks = []
+		for m in MarkdownImporter._sections.finditer(text):
+			heading = m.group(1).strip()
+			heading_breaks.append((heading, m.start(0)))
+
+		sections = []
+		headings = []
+		locations = []
+
+		def add_section(heading, s):
+			s = s.strip()
+			if not s:
+				return
+
+			s = "\n".join(s.split("\n")[2:])
+			paragraphs = s.split("\n\n")
+
+			h_key = len(headings)
+			headings.append(heading)
+			for i, paragraph in enumerate(paragraphs):
+				sections.append(paragraph)
+				locations.append((h_key, i + 1))
+
+		if heading_breaks:
+			first_break = heading_breaks[0][1]
+			if first_break > 0:
+				s = text[:first_break]
+				if s:
+					add_section("", s)
+
+			heading_breaks.append(("", len(text)))
+			for ((heading, s), (_, e)) in zip(heading_breaks, heading_breaks[1:]):
+				add_section(heading, text[s:e])
+		else:
+			add_section("", text)
 
 		md = Metadata(
 			version="1.0",
-			unique_id=unique_id,
-			origin="<string>",
+			origin=origin,
 			author=author,
 			title=title,
-			speakers={})
+			locations={
+				'type': 'markdown',
+				'data': {
+					'headings': headings
+				}
+			})
 
 		return self._make_doc(
-			md, [self._preprocess_text(s)], [], locations, show_progress=show_progress)
+			md, sections, ['heading', 'paragraph'], locations, **kwargs)
